"""
WebSocket服务器
1. 提交json格式，返回也json格式。可在浏览器中用JS调用，或通过其它语言访问
2. 提交pickle格式，返回也json格式。内部调用

使用FastAPI搭建的WebSocket服务
1. pip install uvicorn[standard], 必须指定stanadard，否则安装后不支持websocket服务
2. `/ws/json` 可用其它语言调用
3. `/ws/bytes` 内部调用
4. `/ws/client` `/ws/amdin`, 反弹RPC

反向RPC时偶尔出现无响应的情况，应当是大数据量时导致的不稳定，所以自己实现了分片功能
"""
from datetime import datetime
from typing import List, Any, Dict, Union

from fastapi import WebSocket, WebSocketDisconnect, Depends, Query
from fastapi import status
from fastapi.security.utils import get_authorization_scheme_param
from loguru import logger

from .app_ import app
from ..caller import call, before_call
from ..config import IP_BLOCK, IP_ALLOW, ENABLE_RELAY
from ..model import Format, RspModel
from ..serializer.json_ import obj_to_dict, dict_to_json
from ..serializer.pkl_gzip import deserialize, serialize
from ..utils.check_ import check_ip

# 二进制拆包
BYTES_PER_SEND = 1024 * 32


async def get_current_user(ws: WebSocket, token: Union[str, None] = Query(None)):
    from ..config import AUTH_CHECK, AUTH_TOKENS
    if not AUTH_CHECK:
        return 'anonymous'
    authorization: str = ws.headers.get("Authorization")
    scheme, param = get_authorization_scheme_param(authorization)
    # 如果从头取不到token就从query中取
    if param:
        token = param
    return AUTH_TOKENS.get(token, None)


class ConnectionManager:
    def __init__(self):
        self.active_connections: List[WebSocket] = []

    async def connect(self, ws: WebSocket):
        await ws.accept()
        self.active_connections.append(ws)

    def disconnect(self, ws: WebSocket):
        self.active_connections.remove(ws)

    @staticmethod
    async def send_personal_message(message: str, ws: WebSocket):
        await ws.send_text(message)

    async def broadcast(self, message: str):
        for connection in self.active_connections:
            await connection.send_text(message)


manager = ConnectionManager()


async def _do(ws: WebSocket,
              func: str,

              args: List[Any] = [],
              kwargs: Dict[str, Any] = {},
              fmt: Format = Format.JSON,
              cache_get: bool = True, cache_expire: int = 86400,
              async_remote: bool = True,
              user: str = None,  # 没有用到，用于token认证
              ):
    """实际处理函数"""
    try:
        before_call(ws.client.host, user, func)
        key, buf, data = await call(user, func, args, kwargs, cache_get, cache_expire, async_remote)
    except Exception as e:
        # 主要是处理
        key = type(e).__name__
        # 这里没有缓存，因为这个错误是服务器内部检查
        data = RspModel(status=status.HTTP_401_UNAUTHORIZED,
                        datetime=datetime.now().isoformat(),
                        func=func, args=args, kwargs=kwargs)
        data.type = type(e).__name__
        data.data = repr(e)
        data = data.model_dump() if hasattr(data, 'model_dump') else data.dict()
        buf = serialize(data).read()

    if fmt == Format.PKL_GZIP:
        return buf

    if data is None:
        data = deserialize(buf)

    # DataFrame需要csv格式时走此路径
    if fmt == Format.CSV and data['type'] in ('DataFrame', 'Series'):
        return data['data'].to_csv()

    # JSON格式时，为了方便DataFrame的显示，做一下转换
    data['data'] = obj_to_dict(data['data'])

    return data


@app.websocket("/ws/json")
async def websocket_endpoint_json(websocket: WebSocket, user=Depends(get_current_user)):
    """处理JSON，发送JSON或csv"""
    await manager.connect(websocket)
    try:
        while True:
            req = await websocket.receive_json(mode="text")
            logger.info(req)
            rsp = await _do(websocket, **req, user=user)
            # 部分类型换转json有问题，所以使用特殊的转换函数
            await websocket.send_text(dict_to_json(rsp))
    except WebSocketDisconnect:
        manager.disconnect(websocket)


@app.websocket("/ws/bytes")
async def websocket_endpoint_bytes(websocket: WebSocket, user=Depends(get_current_user)):
    """处理pkl.gz发送pkl.gz"""
    await manager.connect(websocket)
    try:
        while True:
            req = await websocket.receive_bytes()
            bl = deserialize(req)
            req = b''
            while len(req) < bl:
                req += await websocket.receive_bytes()
            req = deserialize(req)
            logger.info(req)
            req['fmt'] = Format.PKL_GZIP
            b = await _do(websocket, **req, user=user)
            bl = len(b)
            await websocket.send_bytes(serialize(bl).read())
            for i in range(0, len(b), BYTES_PER_SEND):
                await websocket.send_bytes(b[i:i + BYTES_PER_SEND])

    except WebSocketDisconnect:
        manager.disconnect(websocket)


# 房间，分存客户端和管理端，中转时用于查找对应的ws
rooms_client: Dict[str, WebSocket] = {}
rooms_admin: Dict[str, WebSocket] = {}


@app.websocket("/ws/client")
async def websocket_endpoint_client(websocket: WebSocket, room: str, user=Depends(get_current_user)):
    """被控端。肉鸡接入后，等待控制端接入，然后转发到肉鸡，然后转发。必须要进入同一房间"""
    await manager.connect(websocket)
    try:
        if not ENABLE_RELAY:
            raise Exception(f'Relay offline')

        if user is None:
            raise Exception(f"Unauthorized")

        rooms_client[room] = websocket
        while True:
            req = await websocket.receive_bytes()
            # 可以将被控端的二进制解码，然后转成json，实现跨语言，这里先不处理
            await rooms_admin[room].send_bytes(req)

    except WebSocketDisconnect:
        manager.disconnect(websocket)
    except Exception as e:
        logger.error(e)
        manager.disconnect(websocket)
        await websocket.close()


@app.websocket("/ws/admin")
async def websocket_endpoint_admin(websocket: WebSocket, room: str, user=Depends(get_current_user)):
    """控制端。等待控制端接入，然后转发"""
    await manager.connect(websocket)
    try:
        # 确保连上admin的用户有权限
        from ..config import IP_CHECK

        if not ENABLE_RELAY:
            raise Exception(f'Relay offline')

        if IP_CHECK:
            from IPy import IP
            # 两张清单数据提前处理，加快处理速度
            __IP_ALLOW__ = {IP(k): v for k, v in IP_ALLOW.items()}
            __IP_BLOCK__ = {IP(k): v for k, v in IP_BLOCK.items()}
            host = IP(websocket.client.host)
            if not check_ip(__IP_ALLOW__, host, False):
                raise Exception(f'IP Not Allowed, {host} not in allowlist')
            if not check_ip(__IP_BLOCK__, host, True):
                raise Exception(f'IP Not Allowed, {host} in blocklist')

        if user is None:
            raise Exception(f"Unauthorized")

        rooms_admin[room] = websocket
        while True:
            req = await websocket.receive_bytes()
            await rooms_client[room].send_bytes(req)

    except WebSocketDisconnect:
        manager.disconnect(websocket)
    except Exception as e:
        logger.error(e)
        manager.disconnect(websocket)
        await websocket.close()



import hashlib
import os
import threading
from datetime import datetime

from revolving_asyncio import to_sync

from ksrpc.serializer.pkl_gzip import serialize


def make_key(func, args, kwargs):
    """生成缓存key"""
    args_kwargs = f'{func}_{repr(args)}_{repr(kwargs)}'
    if len(args_kwargs) > 32:
        args_kwargs = hashlib.md5(args_kwargs.encode('utf-8')).hexdigest()
    return f'{func}_{args_kwargs}'


def simple_call(func_name, args, kwargs):
    """简版API调用。没有各种额外功能"""
    key = make_key(func_name, args, kwargs)

    methods = func_name.split('.')
    module = methods[0]

    # 返回的数据包
    d = dict(status=200,  # status.HTTP_200_OK,
             datetime=datetime.now().isoformat(),  # 加查询时间，缓存中也许可以判断是否过期
             func=func_name,
             args=args,
             kwargs=kwargs)

    methods = methods[1:]

    try:

        try:
            # 当前server目录下文件，用于特别处理
            api = __import__(f'{__package__}.server.{module}', fromlist=['*'])
        except ModuleNotFoundError as e:
            # 导入系统包
            if module == __package__:
                raise Exception(f'Not Allowed to call {__package__}')
            api = __import__(module, fromlist=['*'])

        pid = os.getpid()
        tid = threading.get_ident()
        # 转成字符串，后面可能于做cache的key
        print(f'{pid}:{tid}:\t{func_name}\t{args}\t{kwargs}'[:200])

        func = api
        for method in methods:
            func = getattr(func, method)

        # 可以调用的属性
        if callable(func):
            # 例如os.remove
            type_ = 'function'
            func = to_sync(func)(*args, **kwargs)

        # 结果的类型名
        class_name = func.__class__.__name__

        # 不可调用的直接返回
        if class_name == 'module':
            # 例如os.path，但用法为os.path()
            data = repr(func)
        elif class_name in ('dict_keys', 'dict_values'):
            # 无法序列化，只能强行转换
            # 例如：sys.modules.keys(), 也可以sys.modules.keys.__list()
            data = list(func)
        else:
            # 例如math.pi，但用法math.pi()
            data = func

        d['type'] = type(func).__name__
        d['data'] = data

        # pkl序列化，为了能完整还源
        buf = serialize(d).read()

    except Exception as e:
        d['status'] = 500  # status.HTTP_500_INTERNAL_SERVER_ERROR
        d['type'] = type(e).__name__
        d['data'] = repr(e)

        # 由于错误信息也想缓存，所以这里进行编码
        buf = serialize(d).read()

    return key, buf, d



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Author     :wukan
# @License    :(C) Copyright 2022, wukan
# @Date       :2022-05-16

"""
使用wesockets实现的客户端
1. 默认提交pickle格式，返回也是pickle格式
2. 原接口是异步版，必须使用async with，封装成了同步版with
3. 为了使用方便，还提供了不需with的版本，可直接调用

"""
import asyncio
from datetime import datetime
from urllib.parse import urlparse

from loguru import logger
from revolving_asyncio import to_sync
from websockets import __version__ as websockets_version

from ..model import Format, RspModel
from ..serializer.json_ import dict_to_json, json_to_dict, dict_to_obj
from ..serializer.pkl_gzip import serialize, deserialize
from ..utils.check_ import check_methods
from ..utils.notebook import clear_output

is_old = int(websockets_version.split('.')[0]) < 13

# 二进制拆包
BYTES_PER_SEND = 1024 * 32


def process_response_dict(data):
    # data = deserialize(r)
    if data['status'] == 200:
        return data['data']
    return data


def process_response_json(r):
    # json字符串转成字字典
    data = json_to_dict(r)
    # 特殊类型先还原
    if data.get('type', None) in ('DataFrame', 'Series', 'ndarray'):
        data['data'] = dict_to_obj(data['data'])
    # 成功响应直接返回数据区
    if data['status'] == 200:
        return data['data']
    return data


class WebSocketConnection:
    def __init__(self, url: str, token=None):
        path = urlparse(url).path
        assert path.endswith(('/json', '/bytes', '/client', '/admin')), 'Python语言优先使用bytes，其它语言使用json'

        if path.endswith('/json'):
            self._fmt = Format.JSON
        else:
            self._fmt = Format.PKL_GZIP

        self._url = url
        self._token = token
        self._with = False  # 记录是否用了with

    async def __aenter__(self):
        """异步async with"""
        from websockets import connect

        headers = None if self._token is None else {"Authorization": f"Bearer {self._token}"}
        # 默认是2**20，只有1MB，扩充一下成8MB
        if is_old:
            self._client = connect(self._url, extra_headers=headers, max_size=2 ** 23)
        else:
            self._client = connect(self._url, additional_headers=headers, max_size=2 ** 23)

        self._ws = await self._client.__aenter__()
        self._with = True
        return self

    async def __aexit__(self, exc_type=None, exc_value=None, traceback=None):
        """异步async with"""
        await self._client.__aexit__(exc_type, exc_value, traceback)

    def __enter__(self):
        """同步with"""
        to_sync(self.__aenter__)()
        return self

    def __exit__(self, exc_type=None, exc_val=None, exc_tb=None):
        """同步with"""
        to_sync(self.__aexit__)()

    async def call(self, func, args, kwargs,
                   fmt: Format = Format.PKL_GZIP,
                   cache_get: bool = True, cache_expire: int = 3600,
                   async_remote=True,
                   timeout=30):
        # 还没有执行过with就内部主动执行一次
        if not self._with:
            await self.__aenter__()

        # 如果是JSON格式可以考虑返回CSV
        rsp_fmt = self._fmt

        d = dict(func=func, args=args, kwargs=kwargs,
                 fmt=rsp_fmt.value,
                 cache_get=cache_get, cache_expire=cache_expire, async_remote=async_remote)

        if self._fmt == Format.PKL_GZIP:
            # 二进制格式
            b = serialize(d).read()
            bl = len(b)
            await self._ws.send(serialize(bl).read())
            for i in range(0, len(b), BYTES_PER_SEND):
                await self._ws.send(b[i:i + BYTES_PER_SEND])
        else:
            # json格式
            await self._ws.send(dict_to_json(d))

        # 这里收到的数据是否需要解析一下，如果是二进制的，需要特别处理
        rsp = await asyncio.wait_for(self._ws.recv(), timeout)
        if isinstance(rsp, bytes):
            # 二进制，反序列化后是字典
            bl = deserialize(rsp)
            rsp = b''
            while len(rsp) < bl:
                rsp += await asyncio.wait_for(self._ws.recv(), timeout)
            return process_response_dict(deserialize(rsp))
        else:
            return process_response_json(rsp)

    async def reverse(self, recv_timeout=True, clear_cnt=5):
        """反弹RPC的被控端"""
        from ..caller import call

        recv_count = 0
        _clear_cnt = 0
        while True:
            if recv_timeout:
                # 对第一个数据包长度添加超时机制
                timeout_count = 0
                while True:
                    try:
                        # 等15秒
                        req = await asyncio.wait_for(self._ws.recv(), 30)
                        timeout_count = 0
                        break
                    except asyncio.TimeoutError:
                        timeout_count += 1
                        # print('接收超时')
                        if timeout_count >= 2:
                            # 约60秒
                            # print('接收超时断开')
                            return recv_count
            else:
                # 一直等待，无超时
                req = await self._ws.recv()

            recv_count += 1
            bl = deserialize(req)
            req = b''
            while len(req) < bl:
                req += await self._ws.recv()
            req = deserialize(req)
            # 可能显示太多，需要裁剪一些
            logger.info(repr(req)[:200])
            # 不需要缓存
            req['cache_get'] = False
            req['cache_expire'] = 0
            func = req['func']
            try:
                from ..config import METHODS_ALLOW, METHODS_BLOCK, METHODS_CHECK

                # 反弹RPC权限要进行限制
                if METHODS_CHECK:
                    methods = func.split('.')
                    if not check_methods(METHODS_ALLOW, methods, False):
                        raise Exception(f'Method Not Allowed, {func} not in allowlist')
                    if not check_methods(METHODS_BLOCK, methods, True):
                        raise Exception(f'Method Not Allowed, {func} in blocklist')
                # 每个连上去的都分别使用了自己的用户名，这里不需要用户名即可操作
                user = 'reverse'
                key, buf, data = await call(user, **req)
            except Exception as e:
                # 主要是处理
                key = type(e).__name__
                # 这里没有缓存，因为这个错误是服务器内部检查
                data = RspModel(status=401,  # status.HTTP_401_UNAUTHORIZED,
                                datetime=datetime.now().isoformat(),
                                func=func, args=req['args'], kwargs=req['kwargs'])
                data.type = type(e).__name__
                data.data = repr(e)
                data = data.model_dump() if hasattr(data, 'model_dump') else data.dict()
                buf = serialize(data).read()

            # 释放内存
            del data
            del req
            # 将这里分成两种处理方法
            bl = len(buf)
            print(f'需发送字节数:{bl} ', end='')
            await self._ws.send(serialize(bl).read())
            for i in range(0, len(buf), BYTES_PER_SEND):
                print('>', end='')
                await self._ws.send(buf[i:i + BYTES_PER_SEND])
                print('\b=', end='')
            print(' 发送完成')
            # 释放内存
            del buf
            _clear_cnt += 1
            if _clear_cnt >= clear_cnt:
                _clear_cnt = 0
                clear_output()



def check_ip(ip_list, ip, default):
    """IP检查

    找到了，就用约定的值
    没找到，就用默认值default
    """
    for x, y in ip_list.items():
        # if IP(ip) in IP(x):
        if ip in x:
            # 返回指定值
            return y
    # 没找到，反回默认值
    return default


def check_methods(dict_, list_, default):
    """方法权限检查

    找到了，就用约定的值
    没找到，就用默认值default
    """
    d = dict_
    for x in list_:
        d = d.get(x, default)
        if isinstance(d, bool):
            return d
    return default


def get_quota(dict_, list_, default):
    """方法权限检查

    找到了，就用约定的值
    没找到，就用默认值default
    """
    d = dict_
    for x in list_:
        d = d.get(x, default)
        if isinstance(d, int):
            return d
    return default


if __name__ == "__main__":
    from IPy import IP

    IP_ALLOW = {
        IP('127.0.0.0/8'): True,  # 回环地址
        IP('192.168.0.0/16'): True,  # C类内网
        IP('172.16.0.0/12'): True,  # B类内网
        IP('10.0.0.0/8'): True,  # A类内网
    }
    IP_BLOCK = {
        IP('8.8.8.8/32'): False,
        IP('8.8.4.4/32'): False,
    }
    print(check_ip(IP_ALLOW, IP('127.0.0.1'), False))
    print(check_ip(IP_ALLOW, IP('172.31.0.0'), False))
    print(check_ip(IP_ALLOW, IP('114.114.114.114'), False))
    print(check_ip(IP_BLOCK, IP('8.8.8.8'), True))
    print(check_ip(IP_BLOCK, IP('114.114.114.114'), True))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Author     :wukan
# @License    :(C) Copyright 2022, wukan
# @Date       :2022-05-05

"""
默认的系统内序列化与反序列化
1. 使用内置的pickle进行序列化与反序列化。无法跨语言
2. 使用gzip进行压缩

为了能完整的还原数据，在python语言的客户端中都使用pickle进行序列化
在跨语言的环境中，请使用json等更通用的格式
"""

import gzip
import pickle
from io import BytesIO


# !!! module序列化时会导致系统崩溃，所以不使用dill
# import dill as pickle


def serialize(obj):
    """序列化

    !!! 注意：部分对象无法序列化
    """
    b1 = BytesIO()
    pickle.dump(obj, b1)
    b1.seek(0)
    b2 = BytesIO(gzip.compress(b1.read()))
    return b2


def deserialize(buf):
    """反序列化"""
    b1 = gzip.decompress(buf)
    b2 = BytesIO(b1)
    bb = pickle.load(b2)
    return bb



def is_in_notebook():
    import sys
    return 'ipykernel' in sys.modules


def clear_output():
    """
    clear output for both jupyter notebook and the console
    """
    import os
    os.system('cls' if os.name == 'nt' else 'clear')
    if is_in_notebook():
        from IPython.display import clear_output as clear
        clear()



"""
rqdatac服务转发
1. 屏蔽init功能，由服务端进行统一认证
2. 客户端替换了client, 服务端将请求转到client，统一进行execute

"""
from atexit import register

from loguru import logger
from rqdatac import init, reset
from rqdatac.client import get_client

from ..config import RQ_USERNAME, RQ_PASSWORD

init(username=RQ_USERNAME, password=RQ_PASSWORD)
# 用完后清空，防止被客户端获取
del RQ_USERNAME
del RQ_PASSWORD


@register
def _atexit():
    logger.info("{} {}", __name__, _atexit.__name__)
    reset()


def execute(method, *args, **kwargs):
    return get_client().execute(method, *args, **kwargs)


def init(username=None, password=None, addr=("rqdatad-pro.ricequant.com", 16011), *_, **kwargs):
    # 防止用户调用到了init
    pass



from enum import Enum
from typing import List, Any, Dict

# from fastapi import status
from pydantic import BaseModel


class Format(str, Enum):
    JSON = 'json'  # 请求和响应用Json可以跨语言
    PKL_GZIP = 'pkl.gz'  # pickle格式再压缩，可用于网络传输和缓存，限定客户端为python语言
    CSV = 'csv'  # 响应是DataFrame时，返回CSV格式。其它语言解析时可能比较容易


class RspModel(BaseModel):
    # 状态
    status: int = 200  # status.HTTP_200_OK
    # 发生时间
    datetime: str = 'now'
    # #请求信息
    func: str = 'call'
    args: List[Any] = []
    kwargs: Dict[str, Any] = {}

    # #响应信息
    type: str = ''
    data: Any = None


class ReqModel(BaseModel):
    # #以下在query区，进行提交
    func: str = 'call'
    fmt: Format = Format.CSV
    cache_get: bool = True
    cache_expire: int = 86400

    # #以下在body区使用json格式或二进制格式提交
    args: List[Any] = []
    kwargs: Dict[str, Any] = {}



"""
Wind服务转发
1. 必须先开启Wind客户端再启动此服务
2. 返回的数据做了额外处理，额外抛出异常，用于控制缓存
"""
from functools import wraps

from WindPy import w

# TODO: 是否能在线程中启动，需要测试
# print(w.start())

__path__ = []
__all__ = []


def _post_func(func, name):
    @wraps(func)
    def decorated(*args, **kwargs):
        # print(name)
        # 是否能获取到函数名
        error, data = func(*args, **kwargs)
        # 原始错误上层无法感知，导致可能按用户指定的时间缓存一天
        # 所以特别处理一下，遇到错误抛出异常，这样上层就能识别将缓存缩短成1分钟
        if error != 0:
            raise Exception(data)
        return error, data

    return decorated


def __getattr__(name):
    # 对数据做后处理
    return _post_func(getattr(w, name), name)


def start():
    pass



"""
HTTP服务+frp反代

在某宽上FastAPI 服务可以开启，但客户端连接失败，问题出在uvicorn
通过测试，python -m http.server 能运行，所以决定重写这部分

这次的功能非常简单，只为在某容器中提供HTTP服务，其他权限认证等功能全去除

"""
import os
import socketserver
from http.server import BaseHTTPRequestHandler
from urllib.parse import urlparse, parse_qs

import cgi  # https://pypi.org/project/legacy-cgi/

from ksrpc.caller_simple import simple_call
from ksrpc.config import PORT, HOST
from ksrpc.serializer.pkl_gzip import deserialize


class FileHTTPRequestHandler(BaseHTTPRequestHandler):
    def do_GET(self):
        if self.path == '/':
            self.send_response(200)
            self.send_header("Content-type", "text/html")
            self.end_headers()
            self.wfile.write(b"Hello, this is a simple API!")
        else:
            self.send_response(404)
            self.send_header("Content-type", "text/html")
            self.end_headers()
            self.wfile.write(b"Endpoint not found")

    def do_POST(self):
        if self.path.startswith('/api/file'):
            parsed_path = urlparse(self.path)
            query_params = parse_qs(parsed_path.query)
            func = query_params['func'][0]

            # Parse the form data posted
            form = cgi.FieldStorage(
                fp=self.rfile,
                headers=self.headers,
                environ={'REQUEST_METHOD': 'POST', 'CONTENT_TYPE': self.headers['Content-Type'], }
            )
            file = form['file'].file.read()

            key, buf, data = simple_call(func, **deserialize(file))

            # Begin the response
            self.send_response(200)
            self.send_header("content-type", 'application/octet-stream')
            self.send_header("content-disposition", f'attachment; filename="{key}.pkl.gz"')
            self.end_headers()
            self.wfile.write(buf)
            return


if hasattr(os, "fork"):
    TCPServer = socketserver.ForkingTCPServer
else:
    TCPServer = socketserver.ThreadingTCPServer


def main():
    with TCPServer((HOST, PORT), FileHTTPRequestHandler) as server:
        print(f"Serving on port {PORT}")
        server.serve_forever()



"""
jqdatasdk服务转发
1. jqdatasdk中多线程时会多次登录，实现不重复登录
2. 替换底层
"""
from atexit import register

from jqdatasdk.client import JQDataClient
from loguru import logger


class HackThreadingLocal:
    _instance = None


# 解决多线程时多次登录问题
JQDataClient._threading_local = HackThreadingLocal()

from jqdatasdk import auth as _auth
from jqdatasdk import is_auth as _is_auth
from jqdatasdk import logout as _logout


@register
def _atexit():
    logger.info("{} {}", __name__, _atexit.__name__)
    _logout()


__path__ = []
__all__ = []


def __getattr__(name):
    # 登录认证，争取只做一次
    if not _is_auth():
        from ..config import JQ_USERNAME, JQ_PASSWORD

        _auth(JQ_USERNAME, JQ_PASSWORD)
        # 用完后清空，防止被客户端获取
        del JQ_USERNAME
        del JQ_PASSWORD

    return JQDataClient.instance().__getattr__(name)


def auth(username, password, host=None, port=None):
    pass


def logout():
    pass



# 客户端，用于处理请求
# 需要选择到底是http还是ws，还是nng一类的
import asyncio

from loguru import logger
from revolving_asyncio import to_async, to_sync


class RpcClient:
    """是否优先从缓存中获取"""
    cache_get: bool = True
    """缓存超时时间，秒"""
    cache_expire: int = 86400
    """本地是否暴露为异步，async def函数中才能启用"""
    async_local = False
    """远程是否以异步方式掉用，某些情况下设置成同步才不报错"""
    async_remote = True
    """接收超时时间，秒"""
    recv_timeout = 30

    def __init__(self,
                 module,
                 connection,
                 async_local=False,
                 async_remote=True,
                 ):
        """初始化

        Parameters
        ----------
        module: str
            模块名
        connection: Connection
            连接对象
        async_local: bool
            是否暴露为异步调用，async def函数中才能启用
        async_remote: bool
            远程是否以异步方式掉用，某些情况下设置成同步才不报错
        """
        self._connection = connection
        self._module = module
        self._methods = [module]
        self.async_local = async_local
        self.async_remote = async_remote

    def __getattr__(self, method):
        self._methods.append(method)
        return self

    def __call__(self, *args, **kwargs):
        # 排序，参数顺序统一后，排序生成key便不会浪费了
        kwargs = dict(sorted(kwargs.items()))

        func = '.'.join(self._methods)
        # 用完后得重置，否则第二次用时不正确了
        self._methods = [self._module]
        # 指定外部调用方式是同步还是异步
        if self.async_local:
            f = to_async(self._connection.call)
        else:
            f = to_sync(self._connection.call)

        try:
            # 注意：这里没有指定输入输出格式，只有输入二进制，输出二进制的格式
            # 服务器可以支持多种格式是用于非python客户端
            return f(func, args, kwargs,
                     cache_get=self.cache_get, cache_expire=self.cache_expire, async_remote=self.async_remote,
                     timeout=self.recv_timeout)
        except asyncio.TimeoutError:
            self._connection._with = False  # 重置
            logger.warning(f'{func} timeout, {self.recv_timeout}s')
            raise
        except Exception as e:
            self._connection._with = False  # 重置
            logger.warning(f'{func} error, {e}')
            raise

    def __len__(self):
        # 不知怎么回事，被主动调用了
        return 0



"""
rqdatac客户端替换
1. 将底层的rqdatac.client替换成RpcClient
2. init替换，防止登陆
"""


def hack(client):
    # 内部类就可以了，不用暴露出去
    class HackClient:
        PID = -1

        def __init__(self, c):
            self.client = c

        def execute(self, *args, **kwargs):
            return self.client.execute(*args, **kwargs)

        def reset(self):
            pass

        def info(self):
            pass

        def close(self):
            pass

    def __init(username=None, password=None, addr=("rqdatad-pro.ricequant.com", 16011), *_, **kwargs):
        pass

    # 最简洁的用法，直接将最低层取数据的部分替换
    import rqdatac.client
    rqdatac.client._CLIENT = HackClient(client)
    rqdatac.client.init = __init
    import rqdatac
    rqdatac.init = __init



# !/usr/bin/env python
# -*- coding: utf-8 -*-
# @Author     :wukan
# @License    :(C) Copyright 2022, wukan
# @Date       :2022-05-26
"""
reverse rpc，反弹rpc
如果要调用的服务器在内网，并且由于条件所限，无法安装更多的库
所以这里只需小规模的安装即可

以下代码可以直接复制在Notebook中使用

!!!一定要自行修改房间号!!!

!pip install ksrpc-0.3.3-py3-none-any.whl --user --upgrade
"""

import asyncio

from loguru import logger

from ksrpc.connections.websocket import WebSocketConnection

# 重试次数，防止忘记关闭Notebook内核导致占用资源
RETRY_COUNT = 30
TOKEN = 'secret-token-2'
# 注意：房间号请设置一个复杂的字符串，一定不要与其他用户的冲突，否则数据会乱
URL = 'ws://127.0.0.1:8000/ws/client?room=HA9527'


async def async_main():
    logger.info('强调!!!使用完后一定要停止内核，防止长期占用资源!!!')

    cnt = RETRY_COUNT
    while cnt > 0:  # 重连次数，约1分钟
        logger.info('try to connect. {}', cnt)
        try:
            # 连接被控端地址（带/client），同时还要指定约定的房间号字符串
            async with WebSocketConnection(URL, token=TOKEN) as conn:
                logger.info('connected')
                recv_count = await conn.reverse(recv_timeout=True, clear_cnt=3)
                # 空闲断开时会走到这一步
                logger.info('recv timeout')
                if recv_count == 0:
                    # 内部空闲超时是60秒，为了快速停止，加快超时设定
                    cnt -= 2
                else:
                    # 底层收到过数据包，外层重新计数
                    cnt = RETRY_COUNT
        except Exception as e:
            logger.info(e)
            # 重试不能太块，一直在重试时，试着点击重启内核
            await asyncio.sleep(5)
            cnt -= 1
    logger.info('done')


if __name__ == '__main__':
    # TODO WebSocket+async要加这两句，否则报错
    import revolving_asyncio

    revolving_asyncio.apply()

    asyncio.run(async_main())



# !/usr/bin/env python
# -*- coding: utf-8 -*-
# @Author     :wukan
# @License    :(C) Copyright 2022, wukan
# @Date       :2022-05-22
"""
HTTP/WebSocket服务端入口
"""
import uvicorn

# HTTP服务
from ksrpc.app.http_ import *  # noqa
# WebSocket服务
from ksrpc.app.websocket_ import *  # noqa
from ksrpc.config import HOST, PORT


def main():
    uvicorn.run(app, host=HOST, port=PORT)


if __name__ == "__main__":
    main()
